structure Out : sig

  val prop_to_string : NaturalDeduction.prop -> string

  val term_to_string : ProofsAsPrograms.term -> string

  val print_nd : NaturalDeduction.proof -> unit
  val print_pap : ProofsAsPrograms.proof -> unit
  val print_vau : VerificationsAndUses.proof -> unit
  
end = struct
  (* alias structures *)
  structure ND = NaturalDeduction
  structure PaP = ProofsAsPrograms
  structure VaU = VerificationsAndUses
  (* copy datatypes *)
  datatype atom = datatype ND.atom
  datatype prop = datatype ND.prop
  datatype rule = datatype ND.rule
  datatype term = datatype PaP.term

  (* Natural Deduction *)
  fun rule_to_string rule = case rule of
      Hyp u => "hyp_" ^ u
    | TrueI => "TI"
    | FalseE => "FE"
    | AndI => "^I"
    | AndE1 => "^E1"
    | AndE2 => "^E2"
    | OrI1 => "vI1"
    | OrI2 => "vI2"
    | OrE (u,v) => "vE_" ^ u ^ "_" ^ v
    | ImpI u => ">I_" ^ u
    | ImpE => ">E"
  fun atom_to_string atom = case atom of AA => "A" | AB => "B" | AC => "C" | AD => "D" | AE => "E"
  fun prop_to_string prop = case prop of
      PTrue => "T"
    | PFalse => "F"
    | PAtom a => atom_to_string a
    | PAnd (a,b) => (fn (PAtom _|PTrue|PFalse) => prop_to_string a | _ => "(" ^ prop_to_string a ^ ")") a ^ " ^ " ^ (fn (PAtom _|PTrue|PFalse) => prop_to_string b | _ => "(" ^ prop_to_string b ^ ")") b
    | POr (a,b) => "(" ^ prop_to_string a ^ ") v (" ^ prop_to_string b ^ ")"
    | PImp (a as PImp _,b) => "(" ^ prop_to_string a ^ ") > " ^ prop_to_string b
    | PImp (a,b) => prop_to_string a ^ " > " ^ prop_to_string b
  fun nd_judgment_to_string (ND.JTrue p) = prop_to_string p ^ " true"

  (* Proofs as Programs *)
  fun term_precedence term = case term of
      (Case _ | Lambda _) => 1
    | (Abort _ | Fst _ | Snd _ | Inl _ | Inr _ | App _) => 2
    | (Var _ | Unit | Tuple _) => 3
  fun parens ifleq term = if term_precedence term <= ifleq then "(" ^ term_to_string term ^ ")" else term_to_string term
  and term_to_string term = case term of
      Var u => u
    | Unit => "⟨⟩"
    | Abort e => "abort " ^ parens 2 e
    | Tuple (e1,e2) => "⟨" ^ parens 0 e1 ^ "," ^ parens 0 e2 ^ "⟩"
    | Fst e => "fst " ^ parens 2  e
    | Snd e => "snd " ^ parens 2  e
    | Inl e => "INL " ^ parens 2  e
    | Inr e => "INR " ^ parens 2  e
    | Case (e0,(x,e1),(y,e2)) => "case " ^ parens 0 e0 ^ " of INL x => " ^ parens 1 e1 ^ " | INR x => " ^ parens 1 e2
    | Lambda (x,e) => "λ" ^ x ^ "." ^ parens 0 e
    | App (e1,e2) => parens 1 e1 ^ parens 2 e2
  fun pap_judgment_to_string (PaP.JHasType (t,p)) = term_to_string t ^ " : " ^ prop_to_string p

  (* Verifications and Uses *)
  fun vau_rule_to_string rule = case rule of
      VaU.Hyp u => "hyp_" ^ u
    | VaU.TrueI => "TI"
    | VaU.FalseE => "FE"
    | VaU.AndI => "^I"
    | VaU.AndE1 => "^E1"
    | VaU.AndE2 => "^E2"
    | VaU.OrI1 => "vI1"
    | VaU.OrI2 => "vI2"
    | VaU.OrE (u,v) => "vE_" ^ u ^ "_" ^ v
    | VaU.ImpI u => ">I_" ^ u
    | VaU.ImpE => ">E"
    | VaU.DownUp a => "↓↑" ^ atom_to_string a
  fun vau_judgment_to_string judgment = case judgment of
      VaU.JUp p => prop_to_string p ^ " ↑"
    | VaU.JDown p => prop_to_string p ^ " ↓"

  (* Printers *)
  val print_nd : ND.proof -> unit = PrettyPrinter.pretty_print_proof (rule_to_string,nd_judgment_to_string)
  val print_pap : PaP.proof -> unit  = PrettyPrinter.pretty_print_proof (rule_to_string,pap_judgment_to_string)
  val print_vau : VaU.proof -> unit  = PrettyPrinter.pretty_print_proof (vau_rule_to_string,vau_judgment_to_string)
end